from dataclasses import replace

from bussdcc import Process, ContextProtocol, Event, Message

from .. import message, config


class ConfigProcess(Process):
    name = "config"

    def handle_event(self, ctx: ContextProtocol, evt: Event[Message]) -> None:
        if isinstance(evt.payload, message.ConfigInitialized):
            ctx.state.set("config", evt.payload.config)
        elif isinstance(evt.payload, message.ConfigUpdate):
            ctx.state.set("config", evt.payload.config)
            ctx.emit(message.ConfigChanged())
        elif isinstance(evt.payload, message.BusAdded):
            cfg = ctx.state.get("config")
            cfg.buses[evt.payload.bus] = {
                "type": evt.payload.type_,
                "config": evt.payload.data,
            }
            ctx.emit(message.ConfigChanged())
        elif isinstance(evt.payload, message.BusConfigUpdate):
            cfg = ctx.state.get("config")
            cfg.buses[evt.payload.bus]["config"] = evt.payload.data
            ctx.emit(message.ConfigChanged())
        elif isinstance(evt.payload, message.BusDeleted):
            cfg = ctx.state.get("config")
            if cfg is None:
                return

            if evt.payload.bus in cfg.buses:
                del cfg.buses[evt.payload.bus]
                ctx.emit(message.ConfigChanged())
        elif isinstance(evt.payload, message.DeviceAdded):
            cfg = ctx.state.get("config")
            cfg.devices[evt.payload.device] = {
                "type": evt.payload.type_,
                "config": evt.payload.data,
            }
            ctx.emit(message.ConfigChanged())
        elif isinstance(evt.payload, message.DeviceConfigUpdate):
            cfg = ctx.state.get("config")
            cfg.devices[evt.payload.device]["config"] = evt.payload.data
            ctx.emit(message.ConfigChanged())
        elif isinstance(evt.payload, message.DeviceDeleted):
            cfg = ctx.state.get("config")
            if cfg is None:
                return

            if evt.payload.device in cfg.devices:
                del cfg.devices[evt.payload.device]
                ctx.emit(message.ConfigChanged())

        elif isinstance(evt.payload, message.SettingsUpdate):
            payload = evt.payload
            cfg = ctx.state.get("config")

            if cfg is None:
                ctx.state.set(
                    "config",
                    config.Config(settings=payload.settings),
                )
            else:
                ctx.state.update(
                    "config",
                    lambda cfg: replace(cfg, settings=payload.settings),
                )
            ctx.emit(message.ConfigChanged())



from dataclasses import dataclass

from bussdcc import Message

from ..config import Config, Settings


@dataclass(slots=True, frozen=True)
class ConfigInitialized(Message):
    config: Config


@dataclass(slots=True, frozen=True)
class ConfigUpdate(Message):
    config: Config


@dataclass(slots=True, frozen=True)
class DeviceAdded(Message):
    device: str
    type_: str
    data: dict[str, object]


@dataclass(slots=True, frozen=True)
class DeviceConfigUpdate(Message):
    device: str
    data: dict[str, object]


@dataclass(slots=True, frozen=True)
class DeviceDeleted(Message):
    device: str


@dataclass(slots=True, frozen=True)
class BusAdded(Message):
    bus: str
    type_: str
    data: dict[str, object]


@dataclass(slots=True, frozen=True)
class BusConfigUpdate(Message):
    bus: str
    data: dict[str, object]


@dataclass(slots=True, frozen=True)
class BusDeleted(Message):
    bus: str


@dataclass(slots=True, frozen=True)
class SettingsUpdate(Message):
    settings: Settings


@dataclass(slots=True, frozen=True)
class ConfigChanged(Message):
    pass


@dataclass(slots=True, frozen=True)
class ConfigSaved(Message):
    pass



from typing import Any

from bussdcc import (
    ContextProtocol,
    Service,
    Event,
    Message,
    RuntimeProtocol,
    DeviceProtocol,
)
from bussdcc_hardware.registry import registry

from .. import message


class DeviceReconciler:
    def reconcile(
        self,
        runtime: RuntimeProtocol,
        desired: dict[str, DeviceProtocol[Any]],
    ) -> None:
        desired_ids = set(desired.keys())
        actual_ids = {d.id for d in runtime.devices.list()}

        for dev_id in desired_ids - actual_ids:
            try:
                runtime.devices.attach(desired[dev_id])
            except:
                pass

        for dev_id in actual_ids - desired_ids:
            try:
                runtime.devices.detach(dev_id)
            except:
                pass


class DeviceManagerService(Service):
    name = "device_manager"

    def __init__(self) -> None:
        self.reconciler = DeviceReconciler()

    def handle_event(self, ctx: ContextProtocol, evt: Event[Message]) -> None:
        if isinstance(evt.payload, (message.ConfigInitialized, message.ConfigChanged)):
            cfg = ctx.state.get("config")

            desired = self.reconcile(registry.buses, cfg.buses)
            self.reconciler.reconcile(ctx.runtime, desired)

            desired.update(self.reconcile(registry.devices, cfg.devices))
            self.reconciler.reconcile(ctx.runtime, desired)

    def reconcile(
        self, col: dict[str, Any], devices_cfg: dict[str, Any]
    ) -> dict[str, DeviceProtocol[Any]]:
        desired = {}

        for device_id, spec in devices_cfg.items():
            entry = col.get(spec["type"])

            if not entry or not entry.available:
                continue

            definition = entry.definition
            if not definition:
                continue

            cfg = definition.config_class.from_dict(spec["config"])

            desired[device_id] = definition.driver_class(id=device_id, config=cfg)

        return desired



from bussdcc import Service, ContextProtocol, Event, Message

from ..config import ConfigStore
from .. import message


class ConfigService(Service):
    name = "config"

    def __init__(self, data_dir: str) -> None:
        self._data_dir = data_dir

    def _save_config(self, ctx: ContextProtocol) -> None:
        cfg = ctx.state.get("config")
        if cfg is None:
            return

        self.cs.data = cfg
        self.cs.save()

        ctx.emit(message.ConfigSaved())

    def start(self, ctx: ContextProtocol) -> None:
        self.cs = ConfigStore(f"{self._data_dir}/config.json")
        if self.cs.data is None:
            return

        ctx.emit(message.ConfigInitialized(self.cs.data))

    def handle_event(self, ctx: ContextProtocol, evt: Event[Message]) -> None:
        if isinstance(evt.payload, message.ConfigChanged):
            self._save_config(ctx)

    def stop(self, ctx: ContextProtocol) -> None:
        self._save_config(ctx)



from typing import Any
from flask import Blueprint, render_template, flash, redirect, url_for, request

from bussdcc_framework.interface.web import current_ctx
from bussdcc_framework.codec import load_value, dump_value
from bussdcc_framework.interface.web import formtree
from bussdcc_hardware.registry import registry

from .... import message

bp = Blueprint("device", __name__, url_prefix="/device")


@bp.route("/")
def index() -> Any:
    ctx = current_ctx()
    cfg = ctx.state.get("config")
    buses = cfg.buses
    devices = cfg.devices
    online_status = {dev.id: dev.online for dev in ctx.runtime.devices.list()}
    return render_template(
        "device/index.html", buses=buses, devices=devices, online_status=online_status
    )


@bp.route("/new_device")
def new_device() -> Any:
    devices = registry.devices
    name = request.args.get("name")
    type_ = request.args.get("type_")
    if name and type_:
        registry_entry = registry.devices[type_]
        definition = registry_entry.definition
        if definition is None:
            flash("Device not available", "warning")
            return redirect(url_for("device.index"))

        tree = formtree.build(definition.config_class)
        return render_template(
            "device/new_form.html",
            name=name,
            type=type_,
            tree=tree,
            action=url_for("device.create_device", name=name, type_=type_),
        )
    else:
        return render_template("device/new_device.html", devices=devices)


@bp.route("/new_bus")
def new_bus() -> Any:
    buses = registry.buses
    name = request.args.get("name")
    type_ = request.args.get("type_")
    if name and type_:
        registry_entry = registry.buses[type_]
        definition = registry_entry.definition
        if definition is None:
            flash("Bus not available", "warning")
            return redirect(url_for("device.index"))
        tree = formtree.build(definition.config_class)
        return render_template(
            "device/new_form.html",
            name=name,
            type=type_,
            tree=tree,
            action=url_for("device.create_bus", name=name, type_=type_),
        )
    else:
        return render_template("device/new_bus.html", buses=buses)


@bp.route("/device/show/<id>")
def show_device(id: str) -> Any:
    ctx = current_ctx()
    device = ctx.runtime.devices.get(id)
    if not device:
        flash("Device not found", "warning")
        return redirect(url_for("index"))

    tree = formtree.build(device.config)

    return render_template(
        "device/show_device.html",
        tree=tree,
        device_id=id,
        action=url_for("device.update_device", id=id),
    )


@bp.route("/bus/show/<id>")
def show_bus(id: str) -> Any:
    ctx = current_ctx()
    device = ctx.runtime.devices.get(id)
    if not device:
        flash("Device not found", "warning")
        return redirect(url_for("index"))

    tree = formtree.build(device.config)

    return render_template(
        "device/show_bus.html",
        tree=tree,
        device_id=id,
        action=url_for("device.update_bus", id=id),
    )


@bp.route("/create_bus/<type_>/<name>", methods=["POST"])
def create_bus(type_: str, name: str) -> Any:
    ctx = current_ctx()

    registry_entry = registry.buses[type_]
    definition = registry_entry.definition
    if definition is None:
        flash("Bus not available", "warning")
        return redirect(url_for("device.index"))

    tree = formtree.build(definition.config_class)
    data = formtree.unflatten(tree, request.form)
    cfg = load_value(definition.config_class, data)
    ctx.emit(message.BusAdded(bus=name, type_=type_, data=dump_value(cfg)))

    return redirect(url_for("device.index"))


@bp.route("/create_device/<type_>/<name>", methods=["POST"])
def create_device(type_: str, name: str) -> Any:
    ctx = current_ctx()

    registry_entry = registry.devices[type_]
    definition = registry_entry.definition
    if definition is None:
        flash("Bus not available", "warning")
        return redirect(url_for("device.index"))

    tree = formtree.build(definition.config_class)
    data = formtree.unflatten(tree, request.form)
    cfg = load_value(definition.config_class, data)
    ctx.emit(message.DeviceAdded(device=name, type_=type_, data=dump_value(cfg)))

    return redirect(url_for("device.index"))


@bp.route("/device/update/<id>", methods=["POST"])
def update_device(id: str) -> Any:
    ctx = current_ctx()

    device = ctx.runtime.devices.get(id)
    if not device:
        flash("Device not found", "warning")
        return redirect(url_for("index"))

    tree = formtree.build(device.config)
    data = formtree.unflatten(tree, request.form)
    cfg = load_value(type(device.config), data)
    ctx.emit(message.DeviceConfigUpdate(device=id, data=dump_value(cfg)))

    return redirect(url_for("device.index"))


@bp.route("/bus/update/<id>", methods=["POST"])
def update_bus(id: str) -> Any:
    ctx = current_ctx()

    device = ctx.runtime.devices.get(id)
    if not device:
        flash("Bus not found", "warning")
        return redirect(url_for("index"))

    tree = formtree.build(device.config)
    data = formtree.unflatten(tree, request.form)
    cfg = load_value(type(device.config), data)
    ctx.emit(message.BusConfigUpdate(bus=id, data=dump_value(cfg)))

    return redirect(url_for("device.index"))


@bp.route("/device/delete/<id>", methods=["POST"])
def delete_device(id: str) -> Any:
    ctx = current_ctx()

    device = ctx.runtime.devices.get(id)
    if not device:
        flash("Device not found", "warning")
        return redirect(url_for("device.index"))

    ctx.emit(message.DeviceDeleted(device=id))
    flash(f"Deleted device '{id}'", "success")

    return redirect(url_for("device.index"))


@bp.route("/bus/delete/<id>", methods=["POST"])
def delete_bus(id: str) -> Any:
    ctx = current_ctx()

    device = ctx.runtime.devices.get(id)
    if not device:
        flash("Bus not found", "warning")
        return redirect(url_for("device.index"))

    ctx.emit(message.BusDeleted(bus=id))
    flash(f"Deleted bus '{id}'", "success")

    return redirect(url_for("device.index"))
